@@ -12,13 +12,19 @@
 
 #include <LTKWindow.h>
 
+#include <LTKMenu.h>
+
 #include <LTKWidget.h>
+#include <LTKCompoundWidget.h>
 #include <LTKBox.h>
+#include <LTKButtonDialog.h>
 #include <LTKButton.h>
 #include <LTKCanvas.h>
 #include <LTKDblBufCanvas.h>
 #include <LTKEmpty.h>
+#include <LTKFileReq.h>
 #include <LTKLabel.h>
+#include <LTKList.h>
 #include <LTKScrollbar.h>
 #include <LTKScrollingCanvas.h>
 #include <LTKTextIn.h>
