@@ -24,14 +24,10 @@
 class LTKEmpty: public LTKWidget {
 public:
 
-  //---------- constructors and destructor ----------
+  //---------- constructor ----------
 
   LTKEmpty();
 
-  ~LTKEmpty();
-
-  virtual LTKWidget *copy() { return new LTKEmpty(); }
-
   //---------- layout ----------
 
   virtual void layout1();
