@@ -25,6 +25,7 @@ class LTKWindow;
 class GfxColor;
 class GfxFont;
 class GfxSubpath;
+class TextPage;
 struct RGBColor;
 
 //------------------------------------------------------------------------
@@ -47,9 +48,19 @@ struct BoundingRect {
 };
 
 struct RGBColor {
-  int r, g, b;
+  double r, g, b;
 };
 
+//------------------------------------------------------------------------
+// Parameters
+//------------------------------------------------------------------------
+
+// Install a private colormap.
+extern GBool installCmap;
+
+// Size of RGB color cube.
+extern int rgbCubeSize;
+
 //------------------------------------------------------------------------
 // XOutputFont
 //------------------------------------------------------------------------
@@ -64,10 +75,10 @@ public:
   // Destructor.
   ~XOutputFont();
 
-  // Does this font match the tag, size, and angle?
-  GBool matches(GString *tag1, double m11, double m12, double m21, double m22)
-    { return tag->cmp(tag1) == 0 && mat11 == m11 && mat12 == m12 &&
-	     mat21 == m21 && mat22 == m22; }
+  // Does this font match the ID, size, and angle?
+  GBool matches(Ref id1, double m11, double m12, double m21, double m22)
+    { return id.num == id1.num && id.gen == id1.gen &&
+	     mat11 == m11 && mat12 == m12 && mat21 == m21 && mat22 == m22; }
 
   // Get X font.
   XFontStruct *getXFont() { return xFont; }
@@ -80,7 +91,7 @@ public:
 
 private:
 
-  GString *tag;
+  Ref id;
   double mat11, mat12, mat21, mat22;
   Display *display;
   XFontStruct *xFont;
@@ -152,7 +163,7 @@ public:
   virtual void startPage(int pageNum, GfxState *state);
 
   // End a page.
-  virtual void endPage() {}
+  virtual void endPage();
 
   // Dump page contents to display.
   virtual void dump();
@@ -191,16 +202,32 @@ public:
   virtual void eoClip(GfxState *state);
 
   //----- text drawing
-  virtual void drawChar(GfxState *state, double x, double y, Guchar c);
+  virtual void beginString(GfxState *state, GString *s);
+  virtual void endString(GfxState *state);
+  virtual void drawChar(GfxState *state, double x, double y,
+			double dx, double dy, Guchar c);
 
   //----- image drawing
   virtual void drawImageMask(GfxState *state, Stream *str,
 			     int width, int height, GBool invert,
 			     GBool inlineImg);
   virtual void drawImage(GfxState *state, Stream *str, int width,
-			 int height, GfxColorSpace *colorSpace,
+			 int height, GfxImageColorMap *colorMap,
 			 GBool inlineImg);
 
+  //----- special access
+
+  // Find a string.  If <top> is true, starts looking at <xMin>,<yMin>;
+  // otherwise starts looking at top of page.  If <bottom> is true,
+  // stops looking at <xMax>,<yMax>; otherwise stops looking at bottom
+  // of page.  If found, sets the text bounding rectange and returns
+  // true; otherwise returns false.
+  GBool findText(char *s, GBool top, GBool bottom,
+		 int *xMin, int *yMin, int *xMax, int *yMax);
+
+  // Get the text which is inside the specified rectangle.
+  GString *getText(int xMin, int yMin, int xMax, int yMax);
+
 private:
 
   LTKWindow *win;		// window
@@ -213,6 +240,9 @@ private:
   GC strokeGC;			// GC with stroke color
   GC fillGC;			// GC with fill color
   Region clipRegion;		// clipping region
+  GBool trueColor;		// set if using a TrueColor visual
+  int rMul, gMul, bMul;		// RGB multipliers (for TrueColor)
+  int rShift, gShift, bShift;	// RGB shifts (for TrueColor)
   Gulong			// color cube
     colors[maxRGBCube * maxRGBCube * maxRGBCube];
   int numColors;		// size of color cube
@@ -226,12 +256,13 @@ private:
   XOutputFont *font;		// current font
   XOutputFontCache *fontCache;	// font cache
   XOutputState *save;		// stack of saved states
+  TextPage *text;		// text from the current page
 
   void updateLineAttrs(GfxState *state, GBool updateDash);
   void doFill(GfxState *state, int rule);
   void doClip(GfxState *state, int rule);
   int convertPath(GfxState *state, XPoint **points, int *size,
-		  int *numPoints, int **lengths, GBool fill);
+		  int *numPoints, int **lengths, GBool fillHack);
   void convertSubpath(GfxState *state, GfxSubpath *subpath,
 		      XPoint **points, int *size, int *n);
   void doCurve(XPoint **points, int *size, int *k,
